var nuanria = nuanria || {};
nuanria.Firefox = nuanria.Firefox || {};
nuanria.globals = nuanria.globals || {};

;(function(){
    var _initialized = false;

    /*
     * Kicks off initialization of RIA. This happens
     * for every frame / window in the page.
     */
    nuanria.Firefox.init = function () {
        // Store the start time of the load
        nuanria.globals.t0 = performance.now();
        //nuanria.utils.logPerform("Start init for", document.location.href, "=", this.t0, "ms");
        beginInit();
    };

    /*
     * Waits for background script to become ready and
     * then proceeds with finishInit()
     */
    function beginInit() {
        if (_initialized) {
            // init has already finished
            return;
        }

        // ping background script until it becomes ready
        //console.log("send ping from " + document.location.href);
        browser.runtime.sendMessage(
            {type: "ria-ping"},
            function(response) {
                finishInit();
            }
        );

        // keep pinging until we get a response
        setTimeout(beginInit, 50);
    }

    /*
     * Performs init tasks including creating a pluginApi
     * object that is compatible with RIA NativeMessaging Host,
     * and initializing the RIA controller.
     */
    function finishInit() {
        if (_initialized) {
            // init has already finished
            return;
        }

        var pluginApi = null;

        if (window.top === self) {
            // pluginApi is only needed for main window
            pluginApi =  new nuanria.Firefox.dgnRiaApi();

            window.addEventListener("unload", function(ev) {
                if (pluginApi.isConnected()) {
                    pluginApi.disconnect();
                }
            });
        }

        nuanria.Firefox.controller = new nuanria.Controller(window).init(null, pluginApi);
        _initialized = true;
        //nuanria.utils.logPerform("Finish Init for", document.location.href, "=", performance.now(), "ms");
    }

    /*
     * Implements the plugin API. Relays messages to and from
     * the Native Messaging Host.
     */
    nuanria.Firefox.dgnRiaApi = function() {
        var _connected = false;
        var _jsDelegate = null;

        this.Status_Stopped   = 1;
        this.Status_NoProfile = 2;
        this.Status_Ready     = 3;
        this.Status_Error     = 4;

        this.Log_Perform      = 6;
        this.Log_Error        = 5;
        this.Log_Warn         = 4;
        this.Log_Info         = 3;
        this.Log_Debug        = 2;
        this.Log_Trace        = 1;

        this.isConnected = function() {
            return _connected;
        }

        this.version = function() {
            return "TEST";
        };

        this.engineState = function() {
            // get the value last provided to delegate
            return 3;
        };

        this.connect = function(jsDelegate) {
            _jsDelegate = jsDelegate;
            browser.runtime.onMessage.addListener(handleDelegateRequest);
            browser.runtime.sendMessage({
                type : "ria-remote-message",
                name : "connect",
                data : JSON.stringify({
                        devicePixelRatio: devicePixelRatio
                    })
            });
            _connected = true;
            return _connected;
        };

        this.disconnect = function() {
            browser.runtime.onMessage.removeListener(handleDelegateRequest);
            browser.runtime.sendMessage({type: "ria-remote-message", name: "disconnect"});
            _connected = false;
        };

        this.setDictationActive = function(active, useNatText) {
            browser.runtime.sendMessage({
                type : "ria-remote-message",
                name : "setDictationActive",
                data : {
                    active     : active,
                    useNatText : useNatText
                }
            });
        };

        this.setActiveURL = function (url) {
            browser.runtime.sendMessage({
                type: "ria-remote-message",
                name: "setActiveURL",
                data: {
                    url : url
                }
            });
        };

        this.setSuspendState = function (suspended) {
            browser.runtime.sendMessage({
                type : "ria-remote-message",
                name : "setSuspendState",
                data : {
                    suspended : suspended
                }
            });
        };

        this.onUserInteraction = function(jsonData) {
            browser.runtime.sendMessage({
                type : "ria-remote-message",
                name : "onUserInteraction",
                data : {
                    jsonData : jsonData
                }
            });
        };

        this.log = function(logLevel, message) {
            browser.runtime.sendMessage({
                type : "ria-remote-message",
                name : "log",
                data : {
                    logLevel : logLevel,
                    message  : message
                }
            });
        };

        function handleDelegateRequest(request, sender) {
            if (request.type == "jsDelegate") {
                var fn = _jsDelegate[request.fnName];
                if (fn) {
                    fn(request.jsonData, function(success, jsonData) {
                        browser.runtime.sendMessage({
                            type      : "jsDelegate-response",
                            pageId    : request.pageId,
                            msgId     : request.msgId,
                            success   : success,
                            jsonData  : jsonData
                        });
                    });
                }
            }
        }
    };

}());
